@@ -1,14 +1,10 @@
-#!/usr/bin/env python2
+#!/usr/bin/env python3
 # -*- coding: utf-8 -*-
 #
-# © 2013-15: jbarlow83 from Github (https://github.com/jbarlow83)
+# © 2013-16: jbarlow83 from Github (https://github.com/jbarlow83)
 #
-#
-# Use Leptonica to detect find and remove page skew.  Leptonica uses the method
-# of differential square sums, which its author claim is faster and more robust
-# than the Hough transform used by ImageMagick.
+# Python FFI wrapper for Leptonica library
 
-from __future__ import print_function, absolute_import, division
 import argparse
 import sys
 import os
@@ -44,22 +40,34 @@ class LeptonicaErrorTrap(object):
 
     """
     def __enter__(self):
+        from io import UnsupportedOperation
         self.tmpfile = TemporaryFile()
 
         # Save the old stderr, and redirect stderr to temporary file
-        self.old_stderr_fileno = os.dup(sys.stderr.fileno())
-        os.dup2(self.tmpfile.fileno(), sys.stderr.fileno())
+        sys.stderr.flush()
+        try:
+            self.copy_of_stderr = os.dup(sys.stderr.fileno())
+            os.dup2(self.tmpfile.fileno(), sys.stderr.fileno(),
+                    inheritable=False)
+        except UnsupportedOperation:
+            self.copy_of_stderr = None
         return
 
     def __exit__(self, exc_type, exc_value, traceback):
         # Restore old stderr
-        os.dup2(self.old_stderr_fileno, sys.stderr.fileno())
+        sys.stderr.flush()
+        if self.copy_of_stderr is not None:
+            os.dup2(self.copy_of_stderr, sys.stderr.fileno())
+            os.close(self.copy_of_stderr)
 
         # Get data from tmpfile (in with block to ensure it is closed)
         with self.tmpfile as tmpfile:
             tmpfile.seek(0)  # Cursor will be at end, so move back to beginning
             leptonica_output = tmpfile.read().decode(errors='replace')
 
+        assert self.tmpfile.closed
+        assert not sys.stderr.closed
+
         # If there are Python errors, let them bubble up
         if exc_type:
             logger.warning(leptonica_output)
@@ -104,6 +112,12 @@ class Pix:
 
     Leptonica's reference counting is not threadsafe. This class can be used
     in a threadsafe manner if a Python threading.Lock protects the data.
+
+    This class treats Pix objects as immutable.  All methods return new
+    modified objects.  This allows convenient chaining:
+
+    >>>   Pix.read('filename.jpg').scale((0.5, 0.5)).deskew().show()
+
     """
 
     def __init__(self, pix):
@@ -117,6 +131,22 @@ class Pix:
         else:
             return "<leptonica.Pix image NULL>"
 
+    def _repr_png_(self):
+        """iPython display hook
+
+        returns png version of image
+        """
+
+        data = ffi.new('l_uint8 **')
+        size = ffi.new('size_t *')
+
+        err = lept.pixWriteMemPng(data, size, self._pix, 0)
+        if err != 0:
+            raise LeptonicaIOError("pixWriteMemPng")
+
+        char_data = ffi.cast('char *', data[0])
+        return ffi.buffer(char_data, size[0])[:]
+
     def __getstate__(self):
         data = ffi.new('l_uint32 **')
         size = ffi.new('size_t *')
@@ -142,6 +172,9 @@ class Pix:
             cdata_uint32, len(state['data']))
         Pix.__init__(self, pix)
 
+    def __eq__(self, other):
+        return self.__getstate__() == other.__getstate__()
+
     @property
     def width(self):
         return self._pix.w
@@ -201,18 +234,29 @@ class Pix:
         "Returns a PIL.Image version of this Pix"
         from PIL import Image
 
-        with LeptonicaErrorTrap():
-            pix_swapped = Pix(lept.pixEndianByteSwapNew(self._pix))
+        # Leptonica manages data in words, so it implicitly does an endian
+        # swap.  Tell Pillow about this when it reads the data.
+        pix = self
+        if sys.byteorder == 'little':
+            if self.mode == 'RGB':
+                raw_mode = 'XBGR'
+            elif self.mode == 'RGBA':
+                raw_mode = 'ABGR'
+            elif self.mode == '1':
+                raw_mode = '1;I'
+                pix = Pix(lept.pixEndianByteSwapNew(pix._pix))
+            else:
+                raw_mode = self.mode
+                pix = Pix(lept.pixEndianByteSwapNew(pix._pix))
+        else:
+            raw_mode = self.mode  # no endian swap needed
 
-        size = (pix_swapped._pix.wpl * 4, pix_swapped._pix.h)
-        buf = ffi.buffer(pix_swapped._pix.data, size[0] * size[1])
+        size = (pix._pix.w, pix._pix.h)
+        bytecount = pix._pix.wpl * 4 * pix._pix.h
+        buf = ffi.buffer(pix._pix.data, bytecount)
+        stride = pix._pix.wpl * 4
 
-        im_raw = Image.frombytes(self.mode, size, buf, 'raw')
-
-        # Leptonica stores images in 32-bit words
-        # Need to crop the any trailing amount
-        box = (0, 0, self.width, self.height)
-        im = im_raw.crop(box)
+        im = Image.frombytes(self.mode, size, buf, 'raw', raw_mode, stride)
 
         return im
 
@@ -231,10 +275,10 @@ class Pix:
         with LeptonicaErrorTrap():
             return Pix(lept.pixDeskew(self._pix, reduction_factor))
 
-    def scale(self, scalex, scaley):
+    def scale(self, scale_xy):
         "Returns the pix object rescaled according to the proportions given."
         with LeptonicaErrorTrap():
-            return Pix(lept.pixScale(self._pix, scalex, scaley))
+            return Pix(lept.pixScale(self._pix, scale_xy[0], scale_xy[1]))
 
     def rotate180(self):
         with LeptonicaErrorTrap():
@@ -352,6 +396,33 @@ class Pix:
                 black,
                 white))
 
+    def gamma_trc(self, gamma=1.0, minval=0, maxval=255):
+        with LeptonicaErrorTrap():
+            return Pix(lept.pixGammaTRC(
+                ffi.NULL,
+                self._pix,
+                gamma,
+                minval,
+                maxval
+                ))
+
+    def background_norm(
+            self, mask=None, grayscale=None, tile_size=(10, 15), fg_threshold=60,
+            min_count=40, bg_val=200, smooth_kernel=(2, 1)):
+        with LeptonicaErrorTrap():
+            return Pix(lept.pixBackgroundNorm(
+                self._pix,
+                mask or ffi.NULL,
+                grayscale or ffi.NULL,
+                tile_size[0],
+                tile_size[1],
+                fg_threshold,
+                min_count,
+                bg_val,
+                smooth_kernel[0],
+                smooth_kernel[1]
+                ))
+
     @staticmethod
     @lru_cache(maxsize=1)
     def make_pixel_sum_tab8():
@@ -462,6 +533,22 @@ def deskew(infile, outfile, dpi):
         raise LeptonicaIOError("Failed to open destination file: %s" % outfile)
 
 
+def remove_background(infile, outfile, tile_size=(40, 60), gamma=1.0,
+                      black_threshold=70, white_threshold=190):
+    try:
+        pix = Pix.read(infile)
+    except LeptonicaIOError:
+        raise LeptonicaIOError("Failed to open file: %s" % infile)
+
+    pix = pix.background_norm(tile_size=tile_size).gamma_trc(
+            gamma, black_threshold, white_threshold)
+
+    try:
+        pix.write_implied_format(outfile)
+    except LeptonicaIOError:
+        raise LeptonicaIOError("Failed to open destination file: %s" % outfile)
+
+
 if __name__ == '__main__':
     parser = argparse.ArgumentParser(
         description="Python wrapper to access Leptonica")
