@@ -13,6 +13,7 @@ import PyPDF2 as pypdf
 from ocrmypdf import ExitCode
 from ocrmypdf import leptonica
 from ocrmypdf.pdfa import file_claims_pdfa
+import platform
 
 
 if sys.version_info.major < 3:
@@ -22,11 +23,11 @@ if sys.version_info.major < 3:
 TESTS_ROOT = os.path.abspath(os.path.dirname(__file__))
 SPOOF_PATH = os.path.join(TESTS_ROOT, 'spoof')
 PROJECT_ROOT = os.path.dirname(TESTS_ROOT)
-OCRMYPDF = os.path.join(PROJECT_ROOT, 'OCRmyPDF.sh')
 TEST_RESOURCES = os.path.join(PROJECT_ROOT, 'tests', 'resources')
 TEST_OUTPUT = os.environ.get(
     'OCRMYPDF_TEST_OUTPUT',
     default=os.path.join(PROJECT_ROOT, 'tests', 'output', 'main'))
+OCRMYPDF = [sys.executable, '-m', 'ocrmypdf']
 
 
 def running_in_docker():
@@ -34,6 +35,10 @@ def running_in_docker():
     return os.path.exists('/.dockerinit')
 
 
+def is_linux():
+    return platform.system() == 'Linux'
+
+
 def setup_module():
     with suppress(FileNotFoundError):
         shutil.rmtree(TEST_OUTPUT)
@@ -41,15 +46,6 @@ def setup_module():
         os.makedirs(TEST_OUTPUT)
 
 
-def run_ocrmypdf_sh(input_file, output_file, *args, env=None):
-    sh_args = ['sh', OCRMYPDF] + list(args) + [input_file, output_file]
-    sh = Popen(
-        sh_args, close_fds=True, stdout=PIPE, stderr=PIPE,
-        universal_newlines=True, env=env)
-    out, err = sh.communicate()
-    return sh, out, err
-
-
 def _infile(input_basename):
     return os.path.join(TEST_RESOURCES, input_basename)
 
@@ -59,29 +55,35 @@ def _outfile(output_basename):
 
 
 def check_ocrmypdf(input_basename, output_basename, *args, env=None):
+    "Run ocrmypdf and confirmed that a valid file was created"
     input_file = _infile(input_basename)
     output_file = _outfile(output_basename)
 
-    sh, out, err = run_ocrmypdf_sh(input_file, output_file, *args, env=env)
-    if sh.returncode != 0:
+    p, out, err = run_ocrmypdf(input_basename, output_basename, *args, env=env)
+    print(err)  # ensure py.test collects the output, use -s to view
+    if p.returncode != 0:
         print('stdout\n======')
         print(out)
         print('stderr\n======')
         print(err)
-    assert sh.returncode == 0
+    assert p.returncode == 0
     assert os.path.exists(output_file), "Output file not created"
     assert os.stat(output_file).st_size > 100, "PDF too small or empty"
+    assert out == "", \
+        "The following was written to stdout and should not have been: \n" + \
+        "<stdout>\n" + out + "\n</stdout>"
     return output_file
 
 
-def run_ocrmypdf_env(input_basename, output_basename, *args, env=None):
+def run_ocrmypdf(input_basename, output_basename, *args, env=None):
+    "Run ocrmypdf and let caller deal with results"
     input_file = _infile(input_basename)
     output_file = _outfile(output_basename)
 
     if env is None:
         env = os.environ
 
-    p_args = ['ocrmypdf'] + list(args) + [input_file, output_file]
+    p_args = OCRMYPDF + list(args) + [input_file, output_file]
     p = Popen(
         p_args, close_fds=True, stdout=PIPE, stderr=PIPE,
         universal_newlines=True, env=env)
@@ -89,45 +91,60 @@ def run_ocrmypdf_env(input_basename, output_basename, *args, env=None):
     return p, out, err
 
 
-def spoof(replace_program, with_spoof):
+def spoof(**kwargs):
     """Modify environment variables to override subprocess executables
 
+    spoof(program1='replacement', ...)
+
     Before running any executable, ocrmypdf checks the environment variable
     OCRMYPDF_PROGRAMNAME to override default program name/location, e.g.
     OCRMYPDF_GS redirects from the system path Ghostscript ("gs") to elsewhere.
+
     """
     env = os.environ.copy()
-    spoofer = os.path.join(SPOOF_PATH, with_spoof)
-    if not os.access(spoofer, os.X_OK):
-        os.chmod(spoofer, 0o755)
-    env['OCRMYPDF_' + replace_program.upper()] = spoofer
+
+    for replace_program, with_spoof in kwargs.items():
+        spoofer = os.path.join(SPOOF_PATH, with_spoof)
+        if not os.access(spoofer, os.X_OK):
+            os.chmod(spoofer, 0o755)
+        env['OCRMYPDF_' + replace_program.upper()] = spoofer
     return env
 
 
 @pytest.fixture
 def spoof_tesseract_noop():
-    return spoof('tesseract', 'tesseract_noop.py')
+    return spoof(tesseract='tesseract_noop.py')
 
 
 @pytest.fixture
 def spoof_tesseract_cache():
     if running_in_docker():
         return os.environ.copy()
-    return spoof('tesseract', "tesseract_cache.py")
+    return spoof(tesseract="tesseract_cache.py")
 
 
 @pytest.fixture
 def spoof_tesseract_crash():
-    return spoof('tesseract', 'tesseract_crash.py')
+    return spoof(tesseract='tesseract_crash.py')
 
 
 @pytest.fixture
 def spoof_tesseract_big_image_error():
-    return spoof('tesseract', 'tesseract_big_image_error.py')
+    return spoof(tesseract='tesseract_big_image_error.py')
 
 
-def test_quick(spoof_tesseract_noop):
-    check_ocrmypdf('c02-22.pdf', 'test_quick.pdf', env=spoof_tesseract_noop)
+@pytest.fixture
+def spoof_no_tess_no_pdfa():
+    return spoof(tesseract='tesseract_noop.py', gs='gs_pdfa_failure.py')
+
+
+@pytest.fixture
+def spoof_no_tess_pdfa_warning():
+    return spoof(tesseract='tesseract_noop.py', gs='gs_feature_elision.py')
+
+
+def test_quick(spoof_tesseract_cache):
+    check_ocrmypdf('ccitt.pdf', 'test_quick.pdf', env=spoof_tesseract_cache)
 
 
 def test_deskew(spoof_tesseract_noop):
@@ -164,6 +181,38 @@ def test_clean(spoof_tesseract_noop):
                    env=spoof_tesseract_noop)
 
 
+def test_remove_background(spoof_tesseract_noop):
+    from PIL import Image
+
+    # Ensure the input image does not contain pure white/black
+    im = Image.open(_infile('congress.jpg'))
+    assert im.getextrema() != ((0, 255), (0, 255), (0, 255))
+
+    output_pdf = check_ocrmypdf(
+        'congress.jpg', 'test_remove_bg.pdf', '--remove-background',
+        '--image-dpi', '150',
+        env=spoof_tesseract_noop)
+
+    from ocrmypdf.ghostscript import rasterize_pdf
+    import logging
+    log = logging.getLogger()
+
+    output_png = _outfile('remove_bg.png')
+
+    rasterize_pdf(
+        output_pdf,
+        output_png,
+        xres=100,
+        yres=100,
+        raster_device='png16m',
+        log=log)
+
+
+    # The output image should contain pure white and black
+    im = Image.open(output_png)
+    assert im.getextrema() == ((0, 255), (0, 255), (0, 255))
+
+
 # This will run 5 * 2 * 2 = 20 test cases
 @pytest.mark.parametrize(
     "pdf",
@@ -199,6 +248,9 @@ def test_preserve_metadata(spoof_tesseract_noop, output_type):
     assert pdfa_info['output'] == output_type
 
 
+@pytest.mark.skipif(
+    is_linux() and not running_in_docker(),
+    reason="likely to fail if Linux locale is not configured correctly")
 @pytest.mark.parametrize("output_type", [
     'pdfa', 'pdf'
     ])
@@ -210,7 +262,7 @@ def test_override_metadata(spoof_tesseract_noop, output_type):
     chinese = '孔子'
     high_unicode = 'U+1030C is: 𐌌'
 
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         input_file, output_file,
         '--title', german,
         '--author', chinese,
@@ -255,8 +307,8 @@ def test_oversample(spoof_tesseract_cache, renderer):
 
 
 def test_repeat_ocr():
-    sh, _, _ = run_ocrmypdf_sh('graph_ocred.pdf', 'wontwork.pdf')
-    assert sh.returncode != 0
+    p, _, _ = run_ocrmypdf('graph_ocred.pdf', 'wontwork.pdf')
+    assert p.returncode != 0
 
 
 def test_force_ocr(spoof_tesseract_cache):
@@ -403,6 +455,7 @@ def test_maximum_options(spoof_tesseract_cache, renderer, output_type):
     check_ocrmypdf(
         'multipage.pdf', 'test_multipage%s.pdf' % renderer,
         '-d', '-c', '-i', '-g', '-f', '-k', '--oversample', '300',
+        '--remove-background',
         '--skip-big', '10', '--title', 'Too Many Weird Files',
         '--author', 'py.test', '--pdf-renderer', renderer,
         '--output-type', output_type,
@@ -413,19 +466,19 @@ def test_tesseract_missing_tessdata():
     env = os.environ.copy()
     env['TESSDATA_PREFIX'] = '/tmp'
 
-    p, _, err = run_ocrmypdf_env(
+    p, _, err = run_ocrmypdf(
         'graph_ocred.pdf', 'not_a_pdfa.pdf', '-v', '1', '--skip-text', env=env)
     assert p.returncode == ExitCode.missing_dependency, err
 
 
 def test_invalid_input_pdf():
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'invalid.pdf', 'wont_be_created.pdf')
     assert p.returncode == ExitCode.input_file, err
 
 
 def test_blank_input_pdf():
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'blank.pdf', 'still_blank.pdf')
     assert p.returncode == ExitCode.ok
 
@@ -434,7 +487,7 @@ def test_force_ocr_on_pdf_with_no_images(spoof_tesseract_crash):
     # As a correctness test, make sure that --force-ocr on a PDF with no
     # content still triggers tesseract. If tesseract crashes, then it was
     # called.
-    p, _, err = run_ocrmypdf_env(
+    p, _, err = run_ocrmypdf(
         'blank.pdf', 'wont_be_created.pdf', '--force-ocr',
         env=spoof_tesseract_crash)
     assert p.returncode == ExitCode.child_process_error, err
@@ -442,20 +495,20 @@ def test_force_ocr_on_pdf_with_no_images(spoof_tesseract_crash):
 
 
 def test_french(spoof_tesseract_cache):
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'francais.pdf', 'francais.pdf', '-l', 'fra', env=spoof_tesseract_cache)
     assert p.returncode == ExitCode.ok, \
         "This test may fail if Tesseract language packs are missing"
 
 
 def test_klingon():
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'francais.pdf', 'francais.pdf', '-l', 'klz')
     assert p.returncode == ExitCode.bad_args
 
 
 def test_missing_docinfo(spoof_tesseract_noop):
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'missing_docinfo.pdf', 'missing_docinfo.pdf', '-l', 'eng', '-c',
         env=spoof_tesseract_noop)
     assert p.returncode == ExitCode.ok, err
@@ -474,26 +527,26 @@ def test_uppercase_extension(spoof_tesseract_noop):
 
 def test_input_file_not_found():
     input_file = "does not exist.pdf"
-    sh, out, err = run_ocrmypdf_sh(
+    p, out, err = run_ocrmypdf(
         _infile(input_file),
         _outfile("will not happen.pdf"))
-    assert sh.returncode == ExitCode.input_file
+    assert p.returncode == ExitCode.input_file
     assert (input_file in out or input_file in err)
 
 
 def test_input_file_not_a_pdf():
     input_file = __file__  # Try to OCR this file
-    sh, out, err = run_ocrmypdf_sh(
+    p, out, err = run_ocrmypdf(
         _infile(input_file),
         _outfile("will not happen.pdf"))
-    assert sh.returncode == ExitCode.input_file
+    assert p.returncode == ExitCode.input_file
     assert (input_file in out or input_file in err)
 
 
 def test_qpdf_repair_fails():
     env = os.environ.copy()
     env['OCRMYPDF_QPDF'] = os.path.abspath('./spoof/qpdf_dummy_return2.py')
-    p, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         '-v', '1',
         'c02-22.pdf', 'wont_be_created.pdf', env=env)
     print(out)
@@ -502,7 +555,7 @@ def test_qpdf_repair_fails():
 
 
 def test_encrypted():
-    p, out, err = run_ocrmypdf_env('skew-encrypted.pdf', 'wont_be_created.pdf')
+    p, out, err = run_ocrmypdf('skew-encrypted.pdf', 'wont_be_created.pdf')
     assert p.returncode == ExitCode.input_file
     assert out.find('password')
 
@@ -524,19 +577,19 @@ def test_pagesegmode(renderer, spoof_tesseract_cache):
     'tesseract',
     ])
 def test_tesseract_crash(renderer, spoof_tesseract_crash):
-    sh, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'ccitt.pdf', 'wontwork.pdf', '-v', '1',
         '--pdf-renderer', renderer, env=spoof_tesseract_crash)
-    assert sh.returncode == ExitCode.child_process_error
+    assert p.returncode == ExitCode.child_process_error
     assert not os.path.exists(_outfile('wontwork.pdf'))
     assert "ERROR" in err
 
 
 def test_tesseract_crash_autorotate(spoof_tesseract_crash):
-    sh, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'ccitt.pdf', 'wontwork.pdf',
         '-r', env=spoof_tesseract_crash)
-    assert sh.returncode == ExitCode.child_process_error
+    assert p.returncode == ExitCode.child_process_error
     assert not os.path.exists(_outfile('wontwork.pdf'))
     assert "ERROR" in err
     print(out)
@@ -556,22 +609,22 @@ def test_tesseract_image_too_big(renderer, spoof_tesseract_big_image_error):
 def test_no_unpaper():
     env = os.environ.copy()
     env['OCRMYPDF_UNPAPER'] = os.path.abspath('./spoof/no_unpaper_here.py')
-    sh, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'c02-22.pdf', 'wont_be_created.pdf', '--clean', env=env)
-    assert sh.returncode == ExitCode.missing_dependency
+    assert p.returncode == ExitCode.missing_dependency
 
 
 def test_old_unpaper():
     env = os.environ.copy()
     env['OCRMYPDF_UNPAPER'] = os.path.abspath('./spoof/unpaper_oldversion.py')
-    sh, out, err = run_ocrmypdf_env(
+    p, out, err = run_ocrmypdf(
         'c02-22.pdf', 'wont_be_created.pdf', '--clean', env=env)
-    assert sh.returncode == ExitCode.missing_dependency
+    assert p.returncode == ExitCode.missing_dependency
 
 
 def test_algo4():
-    sh, _, _ = run_ocrmypdf_env('encrypted_algo4.pdf', 'wontwork.pdf')
-    assert sh.returncode == ExitCode.encrypted_pdf
+    p, _, _ = run_ocrmypdf('encrypted_algo4.pdf', 'wontwork.pdf')
+    assert p.returncode == ExitCode.encrypted_pdf
 
 
 @pytest.mark.parametrize('renderer', [
@@ -614,15 +667,91 @@ def test_stdin(spoof_tesseract_noop):
     input_file = _infile('francais.pdf')
     output_file = _outfile('test_stdin.pdf')
 
-    p1_args = ['cat', input_file]
-    p1 = Popen(p1_args, close_fds=True, stdin=DEVNULL, stdout=PIPE)
+    # Runs: ocrmypdf - output.pdf < testfile.pdf
+    with open(input_file, 'rb') as input_stream:
+        p_args = OCRMYPDF + ['-', output_file]
+        p = Popen(
+            p_args, close_fds=True, stdout=PIPE, stderr=PIPE,
+            stdin=input_stream, env=spoof_tesseract_noop)
+        out, err = p.communicate()
 
-    p2_args = ['ocrmypdf', '-', output_file]
-    p2 = Popen(
-        p2_args, close_fds=True, stdout=PIPE, stderr=PIPE,
-        stdin=p1.stdout, env=spoof_tesseract_noop)
-    p1.stdout.close()
-    out, err = p2.communicate()
+        assert p.returncode == ExitCode.ok
 
-    assert p2.returncode == ExitCode.ok
 
+def test_stdout(spoof_tesseract_noop):
+    input_file = _infile('francais.pdf')
+    output_file = _outfile('test_stdout.pdf')
+
+    # Runs: ocrmypdf francais.pdf - > test_stdout.pdf
+    with open(output_file, 'wb') as output_stream:
+        p_args = OCRMYPDF + [input_file, '-']
+        p = Popen(
+            p_args, close_fds=True, stdout=output_stream, stderr=PIPE,
+            stdin=DEVNULL, env=spoof_tesseract_noop)
+        out, err = p.communicate()
+
+        assert p.returncode == ExitCode.ok
+
+    from ocrmypdf import qpdf
+    assert qpdf.check(output_file, log=None)
+
+
+def test_masks(spoof_tesseract_noop):
+    check_ocrmypdf('masks.pdf', 'test_masks.pdf', env=spoof_tesseract_noop)
+
+
+def test_linearized_pdf_and_indirect_object(spoof_tesseract_noop):
+    check_ocrmypdf(
+        'epson.pdf', 'test_epson.pdf',
+        env=spoof_tesseract_noop)
+
+
+def test_rotated_skew_timeout():
+    """This document contains an image that is rotated 90 into place with a
+    /Rotate tag and intentionally skewed by altering the transformation matrix.
+
+    This tests for a bug where the combinatino of preprocessing and a tesseract
+    timeout produced a page whose dimensions did not match the original's.
+    """
+
+    input_file = _infile('rotated_skew.pdf')
+    in_pageinfo = pdf_get_all_pageinfo(input_file)[0]
+
+    assert in_pageinfo['height_pixels'] < in_pageinfo['width_pixels'], \
+        "Expected the input page to be landscape"
+    assert in_pageinfo['rotate'] == 90, "Expected a rotated page"
+
+    out = check_ocrmypdf(
+        'rotated_skew.pdf', 'test_rotated_skew.pdf',
+        '--pdf-renderer', 'hocr',
+        '--deskew', '--tesseract-timeout', '0')
+
+    out_pageinfo = pdf_get_all_pageinfo(out)[0]
+
+    assert out_pageinfo['height_pixels'] > out_pageinfo['width_pixels'], \
+        "Expected the output page to be portrait"
+
+    assert out_pageinfo['rotate'] == 0, \
+        "Expected no page rotation for output"
+
+    assert in_pageinfo['width_pixels'] == out_pageinfo['height_pixels'] and \
+        in_pageinfo['height_pixels'] == out_pageinfo['width_pixels'], \
+        "Expected page rotation to be baked in"
+
+
+def test_ghostscript_pdfa_failure(spoof_no_tess_no_pdfa):
+    p, out, err = run_ocrmypdf(
+        'ccitt.pdf', 'test_pdfa_failure.pdf',
+        env=spoof_no_tess_no_pdfa)
+    assert p.returncode == 4, "Expected return code 4 when PDF/A fails"
+
+
+def test_ghostscript_feature_elision(spoof_no_tess_pdfa_warning):
+    check_ocrmypdf('ccitt.pdf', 'test_feature_elision.pdf',
+                   env=spoof_no_tess_pdfa_warning)
+
+
+def test_very_high_dpi(spoof_tesseract_cache):
+    "Checks for a Decimal quantize error with high DPI, etc"
+    check_ocrmypdf('2400dpi.pdf', 'test_2400dpi.pdf',
+                   env=spoof_tesseract_cache)
